/*
*  Power BI Visual CLI
*
*  Copyright (c) Microsoft Corporation
*  All rights reserved.
*  MIT License
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the ""Software""), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
*  THE SOFTWARE.
*/
"use strict";

import "core-js/stable";
import "../../style/visual.less";
import powerbi from "powerbi-visuals-api";
import VisualConstructorOptions = powerbi.extensibility.visual.VisualConstructorOptions;
import VisualUpdateOptions = powerbi.extensibility.visual.VisualUpdateOptions;
import IVisual = powerbi.extensibility.visual.IVisual;
import EnumerateVisualObjectInstancesOptions = powerbi.EnumerateVisualObjectInstancesOptions;
import VisualObjectInstance = powerbi.VisualObjectInstance;
import DataView = powerbi.DataView;
import VisualObjectInstanceEnumerationObject = powerbi.VisualObjectInstanceEnumerationObject;
import IVisualEventService = powerbi.extensibility.IVisualEventService;
import SelectionManager=powerbi.extensibility.ISelectionManager;
import * as visualInterfaces from "./visualInterfaces";
import IVisualHost = powerbi.extensibility.visual.IVisualHost;
import { VisualSettings } from "../settings/visualSettings";
import {FilterManager,ManagerFactory,SelectorManager,LayoutManager} from "../managers";
import * as d3 from "d3";
import { getObject } from "powerbi-visuals-utils-dataviewutils/lib/dataViewObjects";
import { getMeasureIndexOfRole } from "powerbi-visuals-utils-dataviewutils/lib/dataRoleHelper";
import { selector, json } from "d3";
import { debuglog } from "util";
import { EventEmitter } from "events";
import * as models from 'powerbi-models';
export class Visual implements IVisual {
    /*
    SelectorManager: a custom manager, used to manage dropDownSelector, listSelector, calendarSelector, and so on
    SelectionManager: powerbi.extensibility
    */
    private events: IVisualEventService;
    private settings: VisualSettings;
    private selectionManager:SelectionManager;
    private layoutManager: visualInterfaces.ILayoutManager;
    private selectorManager: visualInterfaces.ISelectorManager&EventEmitter;
    private filterManager:visualInterfaces.IFilterManager;
    private readonly id:number;
    private isFirstUpdate:boolean=true;
    //private selectionManager:powerbi.extensibility.ISelectionManager;
    constructor(options: VisualConstructorOptions) {
        console.debug('Visual constructor start');
        console.debug('options:',options);
        this.events = options.host.eventService;
        if (document) {
            //this.selectionManager =options.host.createSelectionManager();
            let container=d3.select(options.element).append("div").classed("container",true);

            //overload context menu
            let coverContextMenu=false;
            if(coverContextMenu){
                d3.select(options.element).on('contextmenu', () => {
                    const mouseEvent: MouseEvent = <MouseEvent>d3.event;
                    //const eventTarget: EventTarget = mouseEvent.target;
                    //let dataPoint = d3.select(eventTarget).datum();
                    this.selectionManager.showContextMenu({}, {
                        x: mouseEvent.clientX,
                        y: mouseEvent.clientY
                    });
                    //mouseEvent.preventDefault();
                });
            }

            //First, create the filterManager
            this.filterManager=ManagerFactory.CreateFilterManager(FilterManager,options.host);
            //Then, layoutManager and selectorManager, both use the filterManager
            //Lagyoutmanager manage the whole div and the slicer-header
            this.layoutManager=ManagerFactory.CreateLayoutManager(LayoutManager,container,this.filterManager);
            //SelectorManager manage selectors
            let selectorContainer=container.append("div").classed("selector-container",true);
            this.selectorManager=ManagerFactory.CreateSelectorManager(SelectorManager,selectorContainer);

            //events
            this.selectorManager.on(this.selectorManager.stringFieldFilter_event,(selectionValues:string[],applyNow:boolean)=>{
                this.filterManager.setFilter_String(selectionValues,applyNow);
            });
            console.debug('selectorManager stringFieldFilter_event registered');
            //set id
            this.id=Math.random();
        }
        console.debug('end constructor');
    }

    public update(options: VisualUpdateOptions) {
        console.debug('visual update start');
        console.debug('id:',this.id);
        console.debug('options:',options);
        //console.debug('operationKind',options.operationKind.toString());
        this.events.renderingStarted(options);
        this.filterManager.inAdvancedEditMode=options.editMode==powerbi.EditMode.Advanced
        this.settings = Visual.parseSettings(options && options.dataViews && options.dataViews[0]);
        let dataView=options.dataViews[0];
        let field=dataView.categorical.categories[0];
        let messures=dataView.categorical.values;
        let defaultSelect:powerbi.DataViewValueColumn;
        let defaultStart:powerbi.DataViewValueColumn;
        let defaultEnd:powerbi.DataViewValueColumn;
        console.debug('Start map messures');
        messures?.map((messure)=>{
            console.debug("map");
            var role=messure.source.roles;
            if(role['defaultSelect']){
                defaultSelect=messure;
            }else if(role['defaultStart']){
                defaultStart=messure;
            }else if(role['defaultEnd']){
                defaultEnd=messure;
            };
        });
        let width:number=options.viewport.width;
        let height:number=options.viewport.height;
        
        console.debug("start layoutManager updateView");
        this.layoutManager.update(dataView,width,height);

        console.debug("start filterManager updateView");
        this.filterManager.update(field);
        
        console.debug("start selectorManager update");
        this.selectorManager.updateData(field,defaultSelect,defaultStart,defaultEnd);
        this.selectorManager.updateFormat(this.settings,width,height);
        console.debug("previous filter",JSON.stringify(options.jsonFilters));
        /*
        Compare two jsonFilters(filterManager.jsonFilter and options.jsonFilters[0]).
        Use method JSON.stringify.
        1. filterManager.jsonFilter is null, options.jsonFilter[0] is null
            Don't restore
        2. filterManager.jsonFilter is Null, options.jsonFilter[0] is not null
            Restore
        3. filterManager.jsonFilter is not null, options.jsonFilter[0] is null
            Don't restore
        4. filterManager.jsonFilter != options.jsonFilter[0]
            Restore
        5. filterManager.jsonFilter == options.jsonFilter[0]
            Don't restore
        */
        if(options.jsonFilters&&options.jsonFilters[0]&&this.isFirstUpdate){
            console.debug('options.jsonFilters[0] is not null');
            if((!this.filterManager.jsonFilter)||
                (JSON.stringify(this.filterManager.jsonFilter)!=JSON.stringify(options.jsonFilters[0]))){
                    console.debug('Two filter are different');
                    let advancedFilter:models.IAdvancedFilter=options.jsonFilters[0] as models.IAdvancedFilter;
                    console.debug("Cast as advancedFilter:",advancedFilter);
                    if(advancedFilter.conditions[0].value=='SyinpoSlicer_AdvancedEditMode_true'){
                        console.debug('previous filter is created in AdvancedEditMode');
                    }else{
                        //Sync slicers
                        console.debug('Sync slicers init  or restore a bookmark start');
                        this.filterManager.setFilter(advancedFilter,false);
                        let oldSelection:string[]=[];
                        advancedFilter.conditions.map(function(condition,index){
                            oldSelection.push(condition.value as string);
                        });
                        this.selectorManager.select(oldSelection);
                        console.debug('Sync slicers init  or restore a bookmark end');
                    }
            }else{
                console.debug('Two filter are equal');
            }//end if
        }else{
            console.debug('Neednt init a sync slicer or restore a bookmark');
        }
        console.debug('visual update end');
        this.isFirstUpdate=false;
        this.filterManager.applyJsonFilter();
        this.events.renderingFinished(options);
    }
    private static parseSettings(dataView: DataView): VisualSettings {
        return <VisualSettings>VisualSettings.parse(dataView);
    }

    /**
     * This function gets called for each of the objects defined in the capabilities files and allows you to select which of the
     * objects and properties you want to expose to the users in the property pane.
     *
     */
    public enumerateObjectInstances(options: EnumerateVisualObjectInstancesOptions): VisualObjectInstance[] | VisualObjectInstanceEnumerationObject {
        return VisualSettings.enumerateObjectInstances(this.settings || VisualSettings.getDefault(), options);
    }
}